@doc raw"""
meff(corr::Vector{uwreal}, plat::Vector{Int64}; pl::Bool=true, data::Bool=false )     

meff(corr::Corr, plat::Vector{Int64}; pl::Bool=true, data::Bool=false)

Computes effective mass for a given correlator corr at a given plateau plat.
Correlator can be passed as an Corr struct or Vector{uwreal}.
    
The flags pl and data allow to show the plots and return data as an extra result.

```@example
data = read_mesons(path, "G5", "G5")
corr_pp = corr_obs.(data)
m = meff(corr_pp[1], [50, 60], pl=false)
```
"""
function meff(corr::Vector{uwreal}, plat::Vector{Int64}; pl::Bool=true, data::Bool=false, mu::Union{Vector{Float64}, Nothing}=nothing)     
    dim = length(corr)                                                                                    
    aux = 0.5 .* log.((corr[2:dim-2] ./ corr[3:dim-1]).^2)     
    mass = plat_av(aux, plat)                                                                            
    uwerr(mass)                       
    if pl == true
        x = 1:length(aux)
        y = value.(aux)
        dy = err.(aux)
        v = value(mass)
        e = err(mass)
        
        figure()
        fill_between(plat[1]:plat[2], v-e, v+e, color="green", alpha=0.75)
        errorbar(x, y, dy, fmt="x", color="black")
        ylabel(L"$m_\mathrm{eff}$")
        xlabel(L"$x_0$")
    
        if !isnothing(mu)
            title(string(L"$\mu_1 = $", mu[1], L" $\mu_2 = $", mu[2]))
        end
        display(gcf())
    end                                                  
    if data == false                   
        return mass                                        
    else               
        return (mass, aux)     
    end                    
end   
meff(corr::Corr, plat::Vector{Int64}; pl::Bool=true, data::Bool=false) = 
    meff(corr.obs, plat, pl=pl, data=data, mu=corr.mu)

## Decay constants                                                                                       
@doc raw"""
dec_const_pcvc(corr::Vector{uwreal}, plat::Vector{Int64}, m::uwreal, mu::Vector{Float64}, y0::Int64 ; pl::Bool=true, data::Bool=false)meff(corr::Corr, plat::Vector{Int64}; pl::Bool=true, data::Bool=false)

dec_const_pcvc(corr::Corr, plat::Vector{Int64}, m::uwreal; pl::Bool=true, data::Bool=false)

Computes decay constant using the PCVC relation for twisted mass fermions. The decay constant is computed in the plateau plat.
Correlator can be passed as an Corr struct or Vector{uwreal}. If it is passed as a uwreal vector, vector of twisted masses mu and source position y0
must be specified.

The flags pl and data allow to show the plots and return data as an extra result.

```@example
data = read_mesons(path, "G5", "G5")
corr_pp = corr_obs.(data)
m = meff(corr_pp[1], [50, 60], pl=false)
f = dec_const_pcvc(corr_pp[1], [50, 60], m, pl=false)
```
"""
function dec_const_pcvc(corr::Vector{uwreal}, plat::Vector{Int64}, m::uwreal, mu::Vector{Float64}, y0::Int64 ; pl::Bool=true, data::Bool=false)
    """
    compute the decay constant when the source is far from the boundaries
    """
    corr_pp = corr[2:end-1]
    dim = length(corr_pp)
                                                                 
    aux = exp.((collect(1:dim) .- y0 ) .* [m for k in 1:dim])
    R = ((aux .* corr_pp).^2).^0.25
    R_av = plat_av(R, plat)
    f = sqrt(2) * (mu[1] + mu[2]) *R_av / m^1.5
    uwerr(f)
    if pl == true
        uwerr(R_av)
        v = value(R_av)
        e = err(R_av)                                
        uwerr.(R)
        figure()
        fill_between(plat[1]:plat[2], v-e, v+e, color="green", alpha=0.75)                                            
        errorbar(1:length(R), value.(R), err.(R), fmt="x", color="black")
        ylabel(L"$R$")
        xlabel(L"$x_0$")
        title(string(L"$\mu_1 = $", mu[1], L" $\mu_2 = $", mu[2]))
        display(gcf())          
    end 
    if data == false
        return f
    else
        return (f, uwdot(sqrt(2) * (mu[1] + mu[2]) / m^1.5, R))                   
    end 
end 
dec_const_pcvc(corr::Corr, plat::Vector{Int64}, m::uwreal; pl::Bool=true, data::Bool=false) =
    dec_const_pcvc(corr.obs, plat, m, corr.mu, corr.y0, pl=pl, data=data)

#t0
@doc raw"""

comp_t0(Y::YData, plat::Vector{Int64}; L::Int64, pl::Bool=false, rw::Union{Matrix{Float64}, Nothing}=nothing)

comp_t0(Y::Vector{YData}, plat::Vector{Int64}; L::Int64, pl::Bool=false, rw::Union{Vector{Matrix{Float64}}, Nothing}=nothing)

Computes t0 using the energy density of the action Ysl(Yang-Mills action).
t0 is computed in the plateau plat.
A (linear) interpolation in t is performed to find t0.

The flag pl allows to show the plot.

```@example
#Single replica
Y = read_ms(path)
rw = read_ms(path_rw)

t0 = comp_t0(Y, [38, 58], L=32)
t0_r = comp_t0(Y, [38, 58], L=32, rw=rw)

#Two replicas
Y1 = read_ms(path1)
Y2 = read_ms(path2)
rw1 = read_ms(path_rw1)
rw2 = read_ms(path_rw2)

t0 = comp_t0([Y1, Y2], [38, 58], L=32, pl=true)
t0_r = comp_t0(Y, [38, 58], L=32, rw=[rw1, rw2], pl=true)

```
"""
function comp_t0(Y::YData, plat::Vector{Int64}; L::Int64, pl::Bool=false, rw::Union{Matrix{Float64}, Nothing}=nothing)
    Ysl = Y.Ysl
    t = Y.t
    id = Y.id

    Ysl = isnothing(rw) ? Ysl : apply_rw(Ysl, rw)
    xmax = size(Ysl, 2)
    nt0 = t0_guess(t, Ysl, plat, L)
    
    Y_aux = Matrix{uwreal}(undef, xmax, 5)
    for i = 1:xmax
        k = 1
        for j = nt0-2:nt0+2
            Y_aux[i, k] = uwreal(Ysl[:, i, j], id)
            k = k + 1
        end
    end
    x = t[nt0-2:nt0+2]
    t2E = [plat_av(Y_aux[:, j], plat) for j=1:5] .* x.^2 / L^3
    
    @. model(x, p) = p[1]*x^3 + p[2]*x^2 + p[3]*x + p[4]
    par = fit_routine(model, x, t2E, 4)
    fmin(x, p) = model(x, p) .- 0.3
    t0 = root_error(fmin, t[nt0], par)
    if pl
        uwerr(t0)
        uwerr.(t2E)
        v = value.(t2E)
        e = err.(t2E)

        figure()
        errorbar(x, v, e, fmt="x")
        
        errorbar(value(t0), 0.3, xerr=err(t0), fmt="x")
        ylabel(L"$t^2E$")
        xlabel(L"$t/a^2$")
        display(gcf())

        figure()
        plot(collect(1:xmax), value.(Y_aux[:, 3]) * t[nt0]^2 / L^3, "x") 
        fill_between(collect(plat[1]:plat[2]), v[3]+e[3], v[3]-e[3], alpha=0.75, color="green")
        ylabel(L"$t^2E(x0, t)$")
        xlabel(L"$x_0/a$")
        title(string(L"$t/a^2 = $", t[nt0]))
        display(gcf())
        end
    return t0
end

function comp_t0(Y::Vector{YData}, plat::Vector{Int64}; L::Int64, pl::Bool=false, rw::Union{Vector{Matrix{Float64}}, Nothing}=nothing)
    nr = length(Y)
    Ysl = getfield.(Y, :Ysl)
    t = getfield.(Y, :t)
    t = t[1]
    id = getfield.(Y, :id)
    vtr = getfield.(Y, :vtr)
    replica = length.(vtr)

    if !all(id .== id[1])
        println("IDs are not equal")
        return nothing
    end

    Ysl = isnothing(rw) ? Ysl : apply_rw.(Ysl, rw)
    tmp = Ysl[1]
    [tmp = cat(tmp, Ysl[k], dims=1) for k = 2:nr]
    nt0 = t0_guess(t, tmp, plat, L)
    xmax = size(tmp, 2)

    Y_aux = Matrix{uwreal}(undef, xmax, 5)
    for i = 1:xmax
        k = 1
        for j = nt0-2:nt0+2
            Y_aux[i, k] = uwreal(tmp[:, i, j], id[1], replica)
            k = k + 1
        end
    end
    x = t[nt0-2:nt0+2]
    t2E = [plat_av(Y_aux[:, j], plat) for j=1:5] .* x.^2 / L^3

    @. model(x, p) = p[1]*x^3 + p[2]*x^2 + p[3]*x + p[4]
    par = fit_routine(model, x, t2E, 4)
    fmin(x, p) = model(x, p) .- 0.3
    t0 = root_error(fmin, t[nt0], par)
    if pl
        uwerr(t0)
        uwerr.(t2E)
        v = value.(t2E)
        e = err.(t2E)

        figure()
        errorbar(x, v, e, fmt="x")
        
        errorbar(value(t0), 0.3, xerr=err(t0), fmt="x")
        ylabel(L"$t^2E$")
        xlabel(L"$t/a^2$")
        display(gcf())

        figure()
        plot(collect(1:xmax), value.(Y_aux[:, 3]) * t[nt0]^2 / L^3, "x")
        fill_between(collect(plat[1]:plat[2]), v[3]+e[3], v[3]-e[3], alpha=0.75, color="green")
        ylabel(L"$t^2E(x0, t)$")
        xlabel(L"$x_0/a$")
        title(string(L"$t/a^2 = $", t[nt0]))
        display(gcf())
    end
    return t0
end
function t0_guess(t::Vector{Float64}, Ysl::Array{Float64, 3}, plat::Vector{Int64}, L::Int64)
    t2E_ax = t.^2 .* mean(mean(Ysl[:, plat[1]:plat[2], :], dims=2), dims=1)[1, 1, :] / L^3
    t0_aux = minimum(abs.(t2E_ax .- 0.3))
    nt0 = findfirst(x-> abs(x - 0.3) == t0_aux, t2E_ax) #index closest value to t0
    return nt0
end